//
// ********************************************************************
// * DISCLAIMER                                                       *
// *                                                                  *
// * The following disclaimer summarizes all the specific disclaimers *
// * of contributors to this software. The specific disclaimers,which *
// * govern, are listed with their locations in:                      *
// *   http://cern.ch/geant4/license                                  *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.                                                             *
// *                                                                  *
// * This  code  implementation is the  intellectual property  of the *
// * GEANT4 collaboration.                                            *
// * By copying,  distributing  or modifying the Program (or any work *
// * based  on  the Program)  you indicate  your  acceptance of  this *
// * statement, and all its terms.                                    *
// ********************************************************************
//
//
// $Id: G4VGaussianQuadrature.hh,v 1.6 2005/03/15 19:11:35 gcosmo Exp $
// GEANT4 tag $Name: geant4-08-00-patch-01 $
//
// Class description:
//
// Base Class for realisation of numerical methodes for integration of functions
// with signature double f(double) by Gaussian quadrature methods
// Roots of ortogonal polynoms and corresponding weights are calculated based on
// iteration method (by bisection Newton algorithm). Constant values for initial
// approximations were derived from the book: M. Abramowitz, I. Stegun, Handbook
// of mathematical functions, DOVER Publications INC, New York 1965 ; chapters 9,
// 10, and 22 .
//
// ---------------------------- Member data: ----------------------------------
//
//  fFunction  - pointer to the function to be integrated
//  fNumber    - the number of points in fAbscissa and fWeight arrays
//  fAbscissa  - array of abscissas, where function will be evaluated
//  fWeight    - array of corresponding weights
//
//
// ----------------------------------------------------------------------
//
// Auxiliary function which returns the value of std::log(gamma-function(x))
//
// G4double 
// GammaLogarithm(G4double xx)

// ------------------------------------------------------------------------------
//
// History:
//             18.04.97   V.Grichine ( Vladimir.Grichine@cern.ch )

#ifndef G4VGAUSSIANQUADRATURE_HH
#define G4VGAUSSIANQUADRATURE_HH

#include "globals.hh"

typedef G4double (*function)(G4double) ;

class G4VGaussianQuadrature
{
  public:

    explicit G4VGaussianQuadrature( function pFunction ) ;
      // Base constructor

    virtual ~G4VGaussianQuadrature() ;
      // Virtual destructor     

    G4double GetAbscissa(G4int index) const ;
    G4double GetWeight(G4int index) const ;
    G4int    GetNumber() const;
      // Access functions

  protected:

    G4double GammaLogarithm(G4double xx) ;

    //  Data members common for GaussianQuadrature family
    //
    function  fFunction ;
    G4double* fAbscissa ;
    G4double* fWeight ;
    G4int     fNumber ;

  private:

    G4VGaussianQuadrature(const G4VGaussianQuadrature&);
    G4VGaussianQuadrature& operator=(const G4VGaussianQuadrature&);
};

#endif
